/*
 *  protocol.c -- binkp implementation
 *
 *  protocol.c is a part of binkd project
 *
 *  Copyright (C) 1996  Dima Maloff, 5047/13
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version. See COPYING.
 */

/*
 * $Id: protocol.c,v 1.21 1996/12/14 07:07:47 mff Exp mff $
 *
 * $Log: protocol.c,v $
 * Revision 1.21  1996/12/14  07:07:47  mff
 * fflush() in remove_from_spool() now works in all versions. Also
 * added poll restoring and non-secure sessions
 *
 * Revision 1.20  1996/12/09  03:54:28  ufm
 * In NT version _NEEDS_ fflush after putc('~') and before gets(...)
 * May be in other versions this feature also needed? May be...
 * NT version - fixed.
 *
 * Revision 1.19  1996/12/09  03:54:28  mff
 * Removed non-blocking mode in NT version
 *
 * Revision 1.18  1996/12/07  12:14:04  mff
 * max_servers limits the number of server processes/threads running
 *
 * Revision 1.17  1996/12/07  07:02:38  mff
 * Fixed open() without close() when truncating a file
 *
 * Revision 1.16  1996/12/05  04:27:37  mff
 * Fixed "timeout after EOB" & "timeout after BSY" bugs
 *
 * Revision 1.15  1996/12/04  14:48:21  ufm
 * Set nonblocking mode in NT.
 *
 * Revision 1.14  1996/12/04  03:49:21  mff
 * Rewritten removing of a file from spool
 *
 * Revision 1.12  1996/12/03  11:27:33  mff
 * NT port by ufm
 */

#include <sys/types.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>

#if !defined(NT) && !defined(W95)
#include <sys/ioctl.h>
#endif

#ifdef HAVE_IOH
#include <io.h>
#endif

#include "Config.h"
#include "sys.h"
#include "iphdr.h"
#include "tools.h"
#include "readcfg.h"
#include "ftnq.h"
#include "bsy.h"
#include "inbound.h"
#include "protocol.h"
#include "protoco2.h"

/*
 * Sets non-blocking mode for a given socket
 */
void setsockopts (SOCKET s)
{
#if (defined(UNIX) || defined(IBMTCPIP)) && defined(FIONBIO)
  int arg;

  arg = 1;
  if (ioctl (s, FIONBIO, (char *) &arg, sizeof arg) < 0)
    Log (1, "ioctl (FIONBIO): %s", TCPERR ());
#endif

/*
   #if defined(NT)
   u_long arg;

   arg = 1;
   if (ioctlsocket (s, FIONBIO, &arg) < 0)
   Log (1, "ioctl (FIONBIO): %s", TCPERR ());
   #endif
 */

/*
   arg = 32 * 1024;
   if (setsockopt (s, SOL_SOCKET, SO_SNDBUF, (char *) &arg, sizeof arg) < 0)
   Log (1, "setsockopt (SO_SNDBUF): %s", TCPERR ());

   arg = 32 * 1024;
   if (setsockopt (s, SOL_SOCKET, SO_RCVBUF, (char *) &arg, sizeof arg) < 0)
   Log (1, "setsockopt (SO_RCVBUF): %s", TCPERR ());
 */

#if defined(UNIX) || defined(EMX)
  if (fcntl (s, F_SETFL, O_NONBLOCK) == -1)
    Log (1, "fcntl: %s", strerror (errno));
#endif
}

/*
 * Fills <<state>> with initial values, allocates buffers, etc.
 */
static int init_protocol (STATE *state, SOCKET socket, FTN_NODE *to)
{
  memset (state, 0, sizeof (STATE));
  state->ibuf = malloc (MAX_BLKSIZE + 1);
  state->isize = state->osize = -1;
  state->obuf = malloc (MAX_BLKSIZE + 1);
  assert (state->ibuf != 0 && state->obuf != 0);
  state->to = to;
  state->maxflvr = 'h';
  setsockopts (state->s = socket);
  TF_ZERO (&state->in);
  TF_ZERO (&state->out);
  TF_ZERO (&state->flo);
  Log (6, "binkp init done, socket # is %i", state->s);
  return 1;
}

/*
 * Clears protocol buffers and queues, closes files, etc.
 */
static int deinit_protocol (STATE *state)
{
  int i;

  if (state->in.f)
  {
    fclose (state->in.f);
    /* We called and there were still files in transfer -- restore poll */
    if (state->to)
    {
      Log (4, "restoring poll with `%c' flavour", state->maxflvr);
      create_poll (&state->to->fa, state->maxflvr);
    }
  }
  if (state->out.f)
    fclose (state->out.f);
  if (state->flo.f)
    fclose (state->flo.f);
  if (state->ibuf)
    free (state->ibuf);
  if (state->obuf)
    free (state->obuf);
  if (state->msgs)
    free (state->msgs);
  if (state->sent_fls)
    free (state->sent_fls);
  if (state->q)
    q_free (state->q);
  for (i = 0; i < state->nfa; ++i)
    bsy_remove (state->fa + i);
  if (state->fa)
    free (state->fa);
  Log (6, "binkp deinit done...");
  return 0;
}

/*
 * Puts a message to the output msg. queue. These msgs will be send
 * right after the current data block.
 */
static void msg_send2 (STATE *state, t_msg m, char *s1, char *s2)
{
  if (!s1)
    s1 = "";
  if (!s2)
    s2 = "";

  state->msgs = realloc (state->msgs, sizeof (BMSG) * (state->n_msgs + 1));
  assert (state->msgs != 0);
  state->msgs[state->n_msgs].t = m;
  state->msgs[state->n_msgs].sz = strlen (s1) + strlen (s2) + 1;
  state->msgs[state->n_msgs].s =
    malloc (state->msgs[state->n_msgs].sz + 3);

  state->msgs[state->n_msgs].s[0] = (unsigned char) ((state->msgs[state->n_msgs].sz >> 8) | 0x80);
  state->msgs[state->n_msgs].s[1] = (unsigned char) state->msgs[state->n_msgs].sz;
  state->msgs[state->n_msgs].s[2] = m;

  strcpy (state->msgs[state->n_msgs].s + 3, s1);
  strcat (state->msgs[state->n_msgs].s + 3, s2);
  state->msgs[state->n_msgs].sz += 2;

  state->n_msgs++;
}

/*
 * Sends a message using format string
 */
static void msg_sendf (STATE *state, t_msg m, char *s,...)
{
  char msg_text[max (MAXPATHLEN, MAX_NETNAME) + 80];
  va_list ap;

  va_start (ap, s);
#ifdef HAVE_SNPRINTF
  vsnprintf (msg_text, sizeof (msg_text), s, ap);
#else
  vsprintf (msg_text, s, ap);
#endif
  va_end (ap);
  msg_send2 (state, m, msg_text, 0);
}

/*
 * 0=data block header, 1=message(command) header
 */
static int send_hdr (STATE *state, int ismsg, int size)
{
  unsigned char b[BLK_HDR_SIZE];
  int i;

  b[0] = (unsigned char) ((size >> 8) | (0x80 * ismsg));
  b[1] = (unsigned char) size;

  if ((i = send (state->s, b, BLK_HDR_SIZE, 0)) < BLK_HDR_SIZE)
    return 0;
  else
    return 1;
}

/*
 * Sends next msg from the msg queue or next data block
 */
static int send_block (STATE *state)
{
  int i, n;

  /* There are unsent msgs and there is no data partially sent */
  if (state->msgs && state->osize == -1)
  {
    for (i = 0; i < state->n_msgs; ++i)
    {
      if (state->msgs[i].s)
      {
	int no;

	if ((no = send (state->s, state->msgs[i].s, state->msgs[i].sz, 0)) <
	    state->msgs[i].sz)
	{
	  if (no == -1 &&
	      (TCPERRNO == TCPERR_WOULDBLOCK ||
	       TCPERRNO == TCPERR_AGAIN))
	    return 1;
	  Log (1, "send msg: %s", TCPERR ());
	  return 0;
	}
	free (state->msgs[i].s);
	state->msgs[i].s = 0;
      }
    }
    free (state->msgs);
    state->msgs = 0;
    state->n_msgs = 0;
    return 1;
  }

  /* There is a file in transfer */
  if (state->out.f)
  {
    /* No data partially sent, so send header */
    if (state->osize == -1)
    {
      state->osize = min (OBLKSIZE, state->out.size - ftell (state->out.f));
      assert (state->osize >= 0);

      Log (7, "next block to send: %li byte(s)", (long) (state->osize));
      if (!send_hdr (state, 0, state->osize))
      {
	if (TCPERRNO == TCPERR_WOULDBLOCK || TCPERRNO == TCPERR_AGAIN)
	{
	  Log (7, "hdr transfer would block");
	  return 1;
	}
	Log (1, "send_hdr: %s", TCPERR ());
	return 0;
      }
      Log (7, "hdr sent");
    }
    /* Data header was sent already */
    if (state->osize != -1)
    {
      if (state->osize != 0 && state->oread == 0)
      {
	Log (7, "freading %li byte(s)", (long) (state->osize));
	if ((n = fread (state->obuf, 1, state->osize, state->out.f))
	    < state->osize)
	{
	  Log (1, "error reading %s: expected %li, read %li",
	       state->out.path, (long) (state->osize), (long) n);
	  return 0;
	}
	state->oread = 1;
      }

      /* The current file have been sent */
      if (state->osize == 0)
      {
	Log (5, "%s: data sent, waitnig for GOT", state->out.netname);
	fclose (state->out.f);
	state->sent_fls = realloc (state->sent_fls,
				   ++(state->n_sent_fls) * sizeof (TFILE));
	assert (state->sent_fls != 0);
	memcpy (state->sent_fls + state->n_sent_fls - 1,
		&state->out,
		sizeof (TFILE));
	TF_ZERO (&state->out);
	state->osize = -1;
      }
      else
      {
	Log (7, "sending %li bytes", (long) (state->osize));
	if ((n = send (state->s, state->obuf, state->osize, 0)) < state->osize)
	{
	  if (n == -1 && TCPERRNO != TCPERR_WOULDBLOCK &&
	      TCPERRNO != TCPERR_AGAIN)
	  {
	    Log (1, "send: %s", TCPERR ());
	    return 0;
	  }
	}
	if (n == state->osize)
	{
	  state->oread = 0;
	  state->osize = -1;
	  Log (7, "data sent");
	}
	else if (n == -1)
	{
	  Log (7, "data transfer would block");
	}
	else
	{
	  memmove (state->obuf, state->obuf + n, state->osize - n);
	  state->osize -= n;
	  Log (7, "partially sent, %li byte(s) left", state->osize);
	}
      }
    }
  }
  return 1;
}

/*
 * Extends parse_args() by sending error message to the remote
 */
static int parse_msg_args (int ac, char **av, char *s, char *ID, STATE *state)
{
  int rc = parse_args (ac, av, s, ID);

  if (!rc)
    msg_send2 (state, M_ERR, ID, ": cannot parse args");
  return rc;
}

/*
 * Truncates the file if action == 't'
 * Removes the file if action == 'd'
 * Otherwise lefts the file unchanged
 */
static int perform_action (char *path, char action)
{
  if (action == 'd')
  {
    if (unlink (path) == -1)
      Log (1, "%s: cannot unlink: %s", path, strerror (errno));
    else
      Log (4, "unlinked %s", path);
  }
  else if (action == 't')
  {
    int h;

    if ((h = open (path, O_WRONLY | O_TRUNC)) == -1)
      Log (1, "%s: cannot truncate: %s", path, strerror (errno));
    else
    {
      Log (4, "truncated %s", path);
      close (h);
    }
  }
  return 1;
}

/*
 * Marks the file in flopath as sent. (Empty .?lo will be removed)
 * If file == 0 just tryes to unlink flopath.
 * If flopath == 0 performs action on file.
 */
static int remove_from_spool (STATE *state, char *flopath,
			       char *file, char action)
{
  char buf[MAXPATHLEN + 1];
  FILE *flo = 0;
  unsigned long offset, curr_offset;
  int i;
  int seek_flag = 0;			    /* Seek _state->flo.f_ to *
					     * _offset_ after
					     * 
					     * processing */
  int empty_flo_flag = 1;

  if (file)
    Log (5, "removing from spool: %s", file);
  else if (flopath)
    Log (5, "removing flo: %s", flopath);
  else
    Log (1, "internal error in remove_from_spool!");

  if (flopath && *flopath)		    /* A file attached via .?lo */
  {
    if (state->flo.f && !strcmp (state->flo.path, flopath))
    {
      flo = state->flo.f;
      offset = ftell (flo);
      fseek (flo, 0, SEEK_SET);
      seek_flag = 1;
    }
    else
    {
      if ((flo = fopen (flopath, "r+b")) == 0)
      {
	Log (5, "remove_from_spool: %s: %s", flopath, strerror (errno));
	return 0;
      }
    }

    while (!feof (flo))
    {
      curr_offset = ftell (flo);
      if (!fgets (buf, MAXPATHLEN, flo))
	break;
      for (i = 0; !isspace (buf[i]); ++i);
      buf[i] = 0;

      if (file && (!strcmp (file, buf) ||
		 ((*buf == '^' || *buf == '#') && !strcmp (file, buf + 1))))
      {
	clearerr (flo);
	if (fseek (flo, curr_offset, SEEK_SET) == EOF)
	  Log (1, "remove_from_spool: fseek(%s): %s", flopath,
	       strerror (errno));
	else if (putc ('~', flo) == EOF)
	  Log (1, "remove_from_spool: fputc(%s): %s", flopath,
	       strerror (errno));
	fflush (flo);
	fgets (buf, MAXPATHLEN, flo);	    /* The line was marked, now skip
					     * it */
      }
      else if (*buf && *buf != '~')
      {
	empty_flo_flag = 0;
      }
    }
    if (seek_flag)
    {
      fseek (flo, offset, SEEK_SET);
      fflush (flo);
    }
    else
    {
      fclose (flo);
      if (empty_flo_flag)
      {
	if (unlink (flopath) == -1)
	  Log (1, "%s: cannot unlink: %s", flopath, strerror (errno));
	else
	  Log (4, "unlinked %s", flopath);
      }
    }
  }
  if (file)
    perform_action (file, action);
  return 1;
}

/*
 * Removes n-th element from the sent files queue,
 * free()'s the queue if it's empty
 */
static void remove_from_sent_files_queue (STATE *state, int n)
{
  state->sent_fls[n].netname[0] = 0;

  for (n = 0; n < state->n_sent_fls; ++n)
    if (state->sent_fls[n].netname[0])
      return;

  if (n >= state->n_sent_fls)
  {
    free (state->sent_fls);
    state->sent_fls = 0;
    state->n_sent_fls = 0;
  }
}

/*
 * These functions down to recv_block() handle binkp msgs
 */

/*
 * Logs down the M_NUL message data
 */
static int NUL (STATE *state, char *buf, int sz)
{
  char *s;

  Log (3, "%s", s = strquote (buf));
  free (s);
  return 1;
}

/*
 * Handles M_ERR msg from the remote
 */
static int RError (STATE *state, char *buf, int sz)
{
  char *s;

  Log (1, "rerror: %s", s = strquote (buf));
  free (s);
  return 1;
}

static int BSY (STATE *state, char *buf, int sz)
{
  char *s;

  Log (1, "got M_BSY: %s", s = strquote (buf));
  free (s);
  return 0;
}

static int ADR (STATE *state, char *s, int sz)
{
  int i, main_AKA_ok = 0;
  char *w;
  FTN_ADDR fa;
  char szFTNAddr[FTN_ADDR_SZ + 1];

  s[sz] = 0;
  for (i = 1; (w = getwordx (s, i, 0)) != 0; ++i)
  {
    if (!parse_ftnaddress (w, &fa) || !is5D (&fa))
    {
      char *q = strquote (s);

      msg_send2 (state, M_ERR, "Bad address", 0);
      Log (1, "Remote passed bad address: `%s'.", q);
      free (w);
      free (q);
      return 0;
    }

    free (w);
    ftnaddress_to_str (szFTNAddr, &fa);

    if (bsy_add (&fa) || !ftnaddress_cmp (&fa, &state->to->fa))
    {
      Log (2, "addr: %s", szFTNAddr);
      state->fa = realloc (state->fa, sizeof (FTN_ADDR) * ++state->nfa);
      assert (state->fa != 0);
      memcpy (state->fa + state->nfa - 1, &fa, sizeof (FTN_ADDR));
      if (!ftnaddress_cmp (state->fa + state->nfa - 1, &state->to->fa))
	main_AKA_ok = 1;
    }
    else
      Log (2, "addr: %s (n/a or busy)", szFTNAddr);
  }
  if (state->nfa == 0)
  {
    Log (1, "no AKAs in common domains or all AKAs are busy");
    msg_send2 (state, M_BSY, "No AKAs in common domains or all AKAs are busy", 0);
    return 0;
  }
  if (state->to != 0 && main_AKA_ok == 0)
  {
    ftnaddress_to_str (szFTNAddr, &state->to->fa);
    Log (1, "called %s, but remote has no such AKA", szFTNAddr);
    return 0;
  }
  return 1;
}

static int PWD (STATE *state, char *pwd, int sz)
{
  FTN_NODE *n;
  int pwd_ok = 0;
  int i;

  pwd[sz] = 0;
  for (i = 0; i < state->nfa; ++i)
  {
    if ((n = get_node_info (state->fa + i)) != 0)
    {
      if (!strcmp (n->pwd, pwd))
	pwd_ok = 1;
      else if (pwd_ok == 0)
      {
	msg_send2 (state, M_ERR, "Bad password", 0);
	Log (1, "`%s': incorrect password", pwd);
	return 0;
      }
    }
  }

  if (pwd_ok)
  {
    Log (2, "pwd protected inbound session");
    state->state = P_SECURE;
  }
  else
  {
    state->state = P_NONSECURE;
  }
  state->q = q_scan_addrs (0, state->fa, state->nfa);
  msg_send2 (state, M_OK, pwd_ok ? "secure" : "non-secure", 0);
  return 1;
}

static int OK (STATE *state, char *buf, int sz)
{
  state->q = q_scan_addrs (0, state->fa, state->nfa);
  state->state = P_SECURE;
  return 1;
}

/*
 * Handles M_FILE msg from the remote
 * M_FILE args: "%s %li %li %li", filename, size, time, offset
 */
static int start_file_recv (STATE *state, char *args, int sz)
{
  const argc = 4;
  char *argv[4];
  long offset;

  if (parse_msg_args (argc, argv, args, "M_FILE", state))
  {
    if (state->in.f &&			    /* Already receiving smthing */
	strcmp (argv[0], state->in.netname))/* ...a file with another name! */
    {
      fclose (state->in.f);
      TF_ZERO (&state->in);
      Log (1, "receiving of %s interrupted", state->in.netname);
    }
    if (state->in.f == 0)
    {
      state->in.start = time (0);
      strnzcpy (state->in.netname, argv[0], MAX_NETNAME);
      strwipe (state->in.netname);
      state->in.size = atol (argv[1]);
      state->in.time = atol (argv[2]);
    }
    offset = atol (argv[3]);

    if (state->in.f == 0)
    {
      if (inb_test (state->in.netname, state->in.size,
		    state->in.time, state->state))
      {
	Log (2, "already have %s (%li byte(s))",
	     state->in.netname, state->in.size);
	msg_sendf (state, M_GOT, "%s %li %li",
		   state->in.netname,
		   (long) state->in.size,
		   (long) state->in.time);
	return 1;
      }
      else if ((state->in.f = inb_fopen (state->in.netname, state->in.size,
					 state->in.time, state->fa,
					 state->state)) == 0)
      {
	return 0;
      }
    }

    if (offset != ftell (state->in.f))
    {
      Log (2, "have partial %s (%li byte(s))",
	   state->in.netname, (long) ftell (state->in.f));
      msg_sendf (state, M_GET, "%s %li %li %li", state->in.netname,
		 (long) state->in.size, (long) state->in.time,
		 (long) ftell (state->in.f));
      ++state->GET_FILE_balance;
      fclose (state->in.f);
      TF_ZERO (&state->in);
      return 1;
    }
    else if (offset != 0)
      --state->GET_FILE_balance;

    Log (3, "receiving %s (%li byte(s), off %li)",
	 state->in.netname, (long) (state->in.size), (long) offset);

    if (fseek (state->in.f, offset, SEEK_SET) == -1)
    {
      Log (1, "fseek: %s", strerror (errno));
      return 0;
    }
    else
      return 1;
  }
  else
    return 0;
}

/*
 * M_GET message from the remote: Resend file from offset
 * M_GET args: "%s %li %li %li", filename, size, time, offset
 */
static int GET (STATE *state, char *args, int sz)
{
  const argc = 4;
  char *argv[4];
  int i, rc = 0;
  long offset;

  if (parse_msg_args (argc, argv, args, "M_GET", state))
  {
    /* Check if the file was already sent */
    for (i = 0; i < state->n_sent_fls; ++i)
    {
      if (!strcmp (argv[0], state->sent_fls[i].netname) &&
	  state->sent_fls[i].size == atol (argv[1]) &&
	  state->sent_fls[i].time == atol (argv[2]))
      {
	if (state->out.f)
	{
	  fclose (state->out.f);
	  TF_ZERO (&state->out);
	}
	memcpy (&state->out, state->sent_fls + i, sizeof (TFILE));
	if ((state->out.f = fopen (state->out.path, "rb")) == 0)
	{
	  Log (1, "GET: %s: %s", state->out.path, strerror (errno));
	  TF_ZERO (&state->out);
	}
	break;
      }
    }

    if (state->out.f && !strcmp (state->out.netname, argv[0]) &&
	state->out.size == atol (argv[1]) &&
	state->out.time == atol (argv[2]))
    {
      if (fseek (state->out.f, offset = atol (argv[3]), SEEK_SET) == -1)
      {
	Log (1, "GET: error seeking %s to %li: %s",
	     argv[0], offset, strerror (errno));
	rc = 0;
      }
      else
      {
	Log (2, "resending %s from %li", argv[0], offset);
	msg_sendf (state, M_FILE, "%s %li %li %li", state->out.netname,
		   (long) state->out.size, (long) state->out.time, offset);
	rc = 1;
      }
    }
    else
      Log (1, "unexpected M_GET for %s", argv[0]);

    return rc;
  }
  else
    return 0;
}

/*
 * M_SKIP: Remote asks us to skip a file. Only a file currently in
 * transfer will be skipped!
 *
 * M_SKIP args: "%s %li %li", filename, size, time
 */
static int SKIP (STATE *state, char *args, int sz)
{
  const argc = 3;
  char *argv[3];
  int n;

  if (parse_msg_args (argc, argv, args, "M_SKIP", state))
  {
    for (n = 0; n < state->n_sent_fls; ++n)
    {
      if (!strcmp (argv[0], state->sent_fls[n].netname) &&
	  state->sent_fls[n].size == atol (argv[1]) &&
	  state->sent_fls[n].time == atol (argv[2]))
      {
	Log (3, "%s skipped by remote", state->sent_fls[n].netname);
	remove_from_sent_files_queue (state, n);
      }
    }
    if (state->out.f && !strcmp (state->out.netname, argv[0]) &&
	state->out.size == atol (argv[1]) &&
	state->out.time == atol (argv[2]))
    {
      fclose (state->out.f);
      Log (3, "%s skipped by remote", state->out.netname);
      TF_ZERO (&state->out);
    }
    return 1;
  }
  else
    return 0;
}

/*
 * M_GOT args: "%s %li %li", filename, size, time
 */
static int GOT (STATE *state, char *args, int sz)
{
  const argc = 3;
  char *argv[3];
  int n;

  if (parse_msg_args (argc, argv, args, "M_GOT", state))
  {
    if (!strcmp (argv[0], state->out.netname) &&
	state->out.size == atol (argv[1]) &&
	state->out.time == atol (argv[2]))
    {
      Log (2, "remote already has %s", state->out.netname);
      fclose (state->out.f);
      remove_from_spool (state, state->out.flo,
			 state->out.path, state->out.action);
      TF_ZERO (&state->out);
    }
    else
    {
      for (n = 0; n < state->n_sent_fls; ++n)
      {
	if (!strcmp (argv[0], state->sent_fls[n].netname) &&
	    state->sent_fls[n].size == atol (argv[1]) &&
	    state->sent_fls[n].time == atol (argv[2]))
	{
	  Log (2, "sent: %s (%.2f CPS)", state->sent_fls[n].path,
	       (double) (state->sent_fls[n].size) /
	       (time (0) == state->sent_fls[n].start ?
		1 : (time (0) - state->sent_fls[n].start)));
	  remove_from_spool (state, state->sent_fls[n].flo,
			state->sent_fls[n].path, state->sent_fls[n].action);
	  remove_from_sent_files_queue (state, n);
	}
      }
    }
    return 1;
  }
  else
    return 0;
}

static int EOB (STATE *state, char *buf, int sz)
{
  state->remote_EOB = 1;
  Log (5, "received EOB");
  return 1;
}

typedef int command (STATE *state, char *buf, int sz);
command *commands[] =
{
  NUL, ADR, PWD, start_file_recv, OK, EOB, GOT, RError, BSY, GET, SKIP
};

/* Recvs next block, processes msgs or writes down the data from the remote */
static int recv_block (STATE *state)
{
  int no;

  int sz = state->isize == -1 ? BLK_HDR_SIZE : state->isize;

  if (sz == 0)
    no = 0;
  else if ((no = recv (state->s, state->ibuf + state->iread,
		       sz - state->iread, 0)) < 0)
  {
    if (TCPERRNO == TCPERR_AGAIN)
    {
      return 1;
    }
    else
    {
      Log (1, "recv: %s", TCPERR ());
#ifdef RDEBUG
      Log (1, "BufSize: %d", sz - state->iread);
#endif
      return 0;
    }
  }
  state->iread += no;
  /* assert (state->iread <= sz); */
  if (state->iread == sz)
  {
    if (state->isize == -1)		    /* reading block header */
    {
      state->imsg = state->ibuf[0] >> 7;
      state->isize = ((((unsigned char *) state->ibuf)[0] & ~0x80) << 8) +
	((unsigned char *) state->ibuf)[1];
      Log (7, "Recvd hdr: %i (%s)", state->isize, state->imsg ? "msg" : "data");
      if (state->isize == 0)
	goto DoNotEvenTryToRecvZeroLengthBlock;
    }
    else
    {
  DoNotEvenTryToRecvZeroLengthBlock:
      Log (7, "Got block: %i (%s)", state->isize, state->imsg ? "msg" : "data");
      if (state->imsg)
      {
	int rc = 1;

	if ((unsigned) (state->ibuf[0]) > M_MAX)
	  Log (1, "unknown msg type from remote: %u", state->ibuf[0]);
	else
	{
	  state->ibuf[state->isize] = 0;
	  rc = commands[(unsigned) (state->ibuf[0])]
	    (state, state->ibuf + 1, state->isize - 1);
	}

	if (rc == 0)
	  return 0;
      }
      else if (state->in.f)
      {
	if (state->isize != 0 &&
	    fwrite (state->ibuf, state->isize, 1, state->in.f) < 1)
	{
	  Log (1, "write error. Disk full?");
	  return 0;
	}
	if (ftell (state->in.f) == state->in.size)
	{
	  fclose (state->in.f);
	  inb_done (state->in.netname, state->in.size,
		    state->in.time, state->fa, state->state);
	  msg_sendf (state, M_GOT, "%s %li %li",
		     state->in.netname,
		     (long) state->in.size,
		     (long) state->in.time);
	  Log (2, "rcvd: %s (%.2f CPS)", state->in.netname,
	       (double) (state->in.size) /
	       (time (0) == state->in.start ?
		1 : (time (0) - state->in.start)));
	  TF_ZERO (&state->in);
	}
	else if (ftell (state->in.f) > state->in.size)
	{
	  Log (1, "rcvd %li extra bytes!", (long) ftell (state->in.f) - state->in.size);
	}
      }
      else if (state->isize > 0)
      {
	Log (7, "ignoring data block (%li byte(s))", (long) state->isize);
      }
      state->isize = -1;
    }
    state->iread = 0;
  }
  return (no == 0 && sz > 0) ? 0 : 1;
}

static void banner (STATE *state)
{
  char szFTNAddr[FTN_ADDR_SZ + 1];
  char *szAkas;
  int i;
  char szLocalTime[60];
  time_t t;
  struct tm *tm;

  msg_send2 (state, M_NUL, "SYS ", sysname);
  msg_send2 (state, M_NUL, "ZYZ ", sysop);
  msg_send2 (state, M_NUL, "LOC ", location);
  msg_send2 (state, M_NUL, "NDL ", nodeinfo);

  time (&t);
  tm = localtime (&t);
  strftime (szLocalTime, sizeof (szLocalTime), "%Y/%m/%d %H:%M:%S", tm);
  msg_send2 (state, M_NUL, "TIME ", szLocalTime);

  szAkas = malloc (nAddr * (FTN_ADDR_SZ + 1));
  assert (szAkas != 0);
  *szAkas = 0;
  for (i = 0; i < nAddr; ++i)
  {
    ftnaddress_to_str (szFTNAddr, pAddr + i);
    strcat (szAkas, " ");
    strcat (szAkas, szFTNAddr);
  }
  msg_send2 (state, M_ADR, szAkas, 0);
  free (szAkas);

  msg_send2 (state, M_NUL,
	     "VER " MYNAME "/" MYVER
#ifdef OS
	     "/" OS
#endif
	     " " PRTCLNAME "/" PRTCLVER, 0);
  if (state->to)
    msg_send2 (state, M_PWD, state->to->pwd, 0);
}

static void netname (char *s, TFILE *q)
{
  if (q->type == 'm')
  {
    sprintf (s, "%08lx.pkt", rnd ());
  }
  else
  {
    char *z = max (strrchr (q->path, '\\'), strrchr (q->path, '/'));

    z = z ? z + 1 : q->path;
    strnzcpy (s, z, MAX_NETNAME);
  }
}

static int start_file_transfer (STATE *state, FTNQ *file)
{
  struct stat sb;
  char buf[MAXPATHLEN + 1];
  FILE *f;
  int i;
  int action = -1;

  if (state->out.f)
    fclose (state->out.f);
  TF_ZERO (&state->out);		    /* No file in transfer */

  if (state->flo.f == 0)		    /* There is no open .?lo */
  {
    state->maxflvr = MAXFLVR (state->maxflvr, file->flvr);
    /* Try to open the suggested file */
    if ((f = fopen (file->path, (file->type == 'l') ? "r+b" : "rb")) == 0 ||
	fstat (fileno (f), &sb) == -1)
    {
      Log (1, "%s: cannot open: %s", file->path, strerror (errno));
      return 0;
    }
    /* We've opened a .?lo */
    if (file->type == 'l')
    {
      state->flo.action = file->action;
      strcpy (state->flo.path, file->path);
      state->flo.f = f;
    }
  }

  if (state->flo.f != 0)
  {
    strcpy (state->out.flo, state->flo.path);
    while (1)
    {
      if (!fgets (buf, MAXPATHLEN, state->flo.f))
      {
	fclose (state->flo.f);
	state->flo.f = 0;		    /* .?lo closed, remove_from_spool
					     * will now unlink it */
	remove_from_spool (state, state->flo.path, 0, 0);
	TF_ZERO (&state->flo);
	return 0;
      }
      for (i = 0; !isspace (buf[i]); ++i);
      buf[i] = 0;
      switch (*buf)
	{
	  case 0:
	  case '~':
	    continue;
	  case '^':
	    action = 'd';
	    strcpy (state->out.path, buf + 1);
	    break;
	  case '#':
	    action = 't';
	    strcpy (state->out.path, buf + 1);
	    break;
	  default:
	    action = 0;
	    strcpy (state->out.path, buf);
	    break;
	}
      if ((f = fopen (state->out.path, "rb")) == 0 ||
	  fstat (fileno (f), &sb) == -1)
      {
	Log (1, "start_file_transfer: %s: %s",
	     state->out.path, strerror (errno));
	remove_from_spool (state, state->out.flo,
			   state->out.path, state->out.action);
      }
      else
      {
	break;
      }
    }
  }

  if (action == -1)
  {
    strcpy (state->out.path, file->path);
    state->out.flo[0] = 0;
    state->out.action = file->action;
    state->out.type = file->type;
  }
  else
  {
    state->out.action = action;
    state->out.type = 0;
  }
  state->out.f = f;
  state->out.size = sb.st_size;
  state->out.time = sb.st_mtime;
  state->out.start = time (0);
  netname (state->out.netname, &state->out);
  Log (3, "sending %s as %s (%li)",
       state->out.path, state->out.netname, (long) state->out.size);

  msg_sendf (state, M_FILE, "%s %li %li 0",
	     state->out.netname, (long) state->out.size,
	     (long) state->out.time);

  return 1;
}

void protocol (SOCKET socket, FTN_NODE *to)
{
  extern int n_servers;
  STATE state;
  struct timeval tv;
  fd_set r, w;
  int no;

  if (!init_protocol (&state, socket, to))
    return;

  tv.tv_sec = nettimeout;		    /* Set up timeout for select() */
  tv.tv_usec = 0;

  banner (&state);
  if (n_servers > max_servers)
  {
    Log (1, "Too many servers");
    msg_send2 (&state, M_BSY, "Too many servers", 0);
  }
  else
  {
    while (1)
    {
      /* If the queue is not empty and there is no file in tranafer */
      if (state.q && state.out.f == 0)
      {
	FTNQ *q;

	while (1)
	{				    /* Next .pkt, .flo or a file */
	  q = 0;
	  if (state.flo.f ||
	      (q = select_next_file (state.q, state.fa, state.nfa)) != 0)
	  {
	    if (start_file_transfer (&state, q))
	      break;
	  }
	  else
	  {
	    q_free (state.q);
	    state.q = 0;
	    break;
	  }
	}
      }

      if (!state.q && !state.local_EOB && state.state == P_SECURE)
      {
	state.local_EOB = 1;
	msg_send2 (&state, M_EOB, 0, 0);
	Log (5, "sent EOB");
      }

      FD_ZERO (&r);
      FD_ZERO (&w);
      FD_SET (socket, &r);
      if (state.msgs || state.out.f)
	FD_SET (socket, &w);

      if (state.local_EOB && state.remote_EOB && state.sent_fls == 0 &&
	  state.GET_FILE_balance == 0 && state.in.f == 0)
      {
	Log (3, "done");
	break;
      }

      if ((no = select (socket + 1, &r, &w, 0, &tv)) == 0)
      {
	Log (1, "timeout!");
	break;
      }
      else if (no < 0)
      {
	Log (1, "select: %s (args: %i %i)", TCPERR (), socket, tv.tv_sec);
	break;
      }
      if (FD_ISSET (socket, &r))	    /* Have something to read */
      {
	if (!recv_block (&state))
	  break;
      }
      if (FD_ISSET (socket, &w))	    /* Clear to send */
      {
	if (!send_block (&state))
	  break;
      }
    }
  }

  if (state.msgs)
    send_block (&state);

  Log (4, "session closed, quitting...");
  deinit_protocol (&state);
}
